<HTML>

    <HEAD>
        <TITLE>UnitTest - DATABUS</TITLE>
    </HEAD>

    <BODY>

        <h2><condata source="databus" key="sysName" />: ONLINE</h2>
        <hr>
        &nbsp;<br>
        <b>This page shows the proper function of the HTTP-DATABUS backend.</b><br>
        It is only used for unit testing and should be removed for production.<br>
        &nbsp;<br>

    </BODY>

</HTML>
